(function(win, doc){

    var QMUIEditor = win.QMUIEditor = {};

    /**
     ********
     * 配置
     ********
     */
    var config = {
        // 段落标签，p或者div
        defaultParagraph: 'p',
        // 字号档次
        fontSizeSmall: 2,
        fontSizeNormal: 3,
        fontSizeBig: 4,
        fontSizeLarge: 5,
        // 标题标签与字体大小设置互斥
        ignoreFontsInTitle: true,
        /**
         * 通信方式
         * @param  {String}   funcName  要调用的java方法名
         * @param  {String}   data      传给java的数据
         */
        onPushNative: function(funcName, data){
            var projScheme = 'qmuieditor://';
            var msg = projScheme + funcName + '/' + data;
            console.log(msg);
        }

    };

    /**
     ********
     * CONST: 常量
     ********
     */
    var CONST = {
        titleTagArr: ['h1','h2','h3','h4','h5','h6'],
        logTag: {
            warning: 'QMUIRICHEDITOR_LOG_WARNING: ',
            error: 'QMUIRICHEDITOR_LOG_ERROR: ',
        }
    };

    /**
     ********
     * state: 编辑器光标位置状态管理
     ********
     */
    var state = QMUIEditor.state = {};
    /**
     * 回调给java的状态集合
     */
    state.map = {
        bold: 'BOLD', // 加粗文字
        italic: 'ITALIC', // 斜体文字
        underline: 'UNDERLINE', // 下划线
        strikeThrough: 'STRIKETHROUGH', // 删除线
        insertUnorderedList: 'UNORDEREDLIST', // 无序列表
        insertOrderedList: 'ORDEREDLIST', // 有序列表
        justifyCenter: 'JUSTIFYCENTER', // 居中
        justifyRight: 'JUSTIFYRIGHT', // 右对齐
        justifyFull: 'JUSTIFYFULL', // 两端对齐
        color: {
            // 字体颜色
            hook: function(output){
                var colorArr = getCommandValue('forecolor').replace("rgb(",'').replace(")",'').replace(/\s/g,'').split(',');
                output.push('COLOR_'+rgb2hex(colorArr[0], colorArr[1], colorArr[2]).toUpperCase());
            }
        },
        fontSize: {
            // 字号档次
            hook: function(output){
                function getSizeName(size){
                    switch(size){
                        case config.fontSizeLarge: return 'LARGE';
                        case config.fontSizeBig: return 'BIG';
                        case config.fontSizeNormal: return 'NORMAL';
                        case config.fontSizeSmall: return 'SMALL';
                        default: return 'UNKNOWN';
                    }
                }
                var size = parseInt(getCommandValue('fontSize'));
                output.push('FONTSIZE_'+getSizeName(size));
            }
        },
        formatBlock: {
            // 段落格式
            hook: function(output){
                var formatBlock = getCommandValue('formatBlock')
                if(formatBlock.length > 0){
                    output.push('FORMATBLOCK_' + formatBlock.toUpperCase())
                }
            }
        },
        selectionContent: {
            // 选中的内容（如果有）
            hook: function(output){
                if (selection.current.length > 0) {
                    output.push('SELECTIONCONTENT_' + win.getSelection().toString());
                }
            }
        },
        link: {
            // 光标所在的链接（如果有）
            hook: function(output){
                if(selection.current.link){
                    output.push('LINK');
                }
            }
        }
    }
    /**
     * 当前状态
     */
    state.current = '';
    /**
     * 提交状态的方法
     */
    state.push = function() {
        var self = this,
            states = [];
        // 遍历map中的状态进行判断
        for(var key in self.map){
            if(Object.prototype.hasOwnProperty.call(self.map, key)){
                var val = self.map[key]
                if(val.hook){
                    val.hook(states)
                }else{
                    getCommandState(key) && states.push(val);
                }
            }
        }
        // 判断当前编辑器内容是否为空
        if(editor.isEmpty()){
            states.push('EMPTY')
        }
        states = states.join('&');
        if(self.current === states){
            return;
        }
        self.current = states;
        config.onPushNative('updateState', states);
    }


    /**
     ********
     * selection: 光标选区相关功能
     ********
     */
    var selection = QMUIEditor.selection = {};
    /**
     * 记录当前选区信息包括：选区起点、选区终点、光标是否处于超链接内（返回具体的超链接dom）、选区内容长度
     */
    selection.current = {
        'link': null,
        'startContainer': null,
        'startOffset': 0,
        'endContainer': null,
        'endOffset': 0,
        'length': 0,
        'selectedText': ''
    };
    /**
     * 备份当前光标选区
     */
    selection.backup = function(){
        var sel = win.getSelection();
        if (sel.rangeCount > 0) {
            var range = sel.getRangeAt(0);
            this.current = {
                'link': range.endContainer.getParentElementByTagName('a') && range.startContainer.getParentElementByTagName('a'),
                'startContainer': range.startContainer,
                'startOffset': range.startOffset,
                'endContainer': range.endContainer,
                'endOffset': range.endOffset,
                'length': sel.toString().length,
                'selectedText': sel.toString()
            };
        }
    }
    /**
     * 将selection.current还原到编辑器里
     */
    selection.restore = function(){
        if(this.current.startContainer && this.current.endContainer){
            var sel = win.getSelection();
            sel.removeAllRanges();
            var range = doc.createRange();
            range.setStart(this.current.startContainer, this.current.startOffset);
            range.setEnd(this.current.endContainer, this.current.endOffset);
            sel.addRange(range);
        }
    }
    /**
     * 取得光标所形成的rect在y方向的数据，包括height,top,bottom:
     * top: rect上边缘到window上边缘的距离
     * bottom: rect下边缘到window上边缘的距离
     * height: rect高度
     * inspired by http://stackoverflow.com/questions/11126047/find-y-coordinate-for-cursor-position-in-div-in-uiwebview#answer-12300505
     */
    selection.getYCoods = function(){
        var sel = win.getSelection(),
            range,
            top = 0,
            bottom = 0,
            height = 0;
        if (sel.rangeCount) {
            range = sel.getRangeAt(0).cloneRange();
            var rect;
            if (range.getClientRects().length > 0) {
                range.collapse(true);
                rect = range.getClientRects()[0];
                top = rect.top;
                height = rect.height;
                bottom = rect.bottom;
            } else {
                var currentNode = range.endContainer,
                    currentOffset = range.endOffset,
                    children = Array.prototype.slice.call(currentNode.childNodes);
                if(children[currentOffset - 1]){
                    rect = children[currentOffset - 1].getBoundingClientRect();
                }else{
                    rect = currentNode.getBoundingClientRect();
                }
                top = rect.top;
                height = rect.height;
                bottom = rect.bottom;
            }
        }
        return {
            top: top,
            bottom: bottom,
            height: height
        };
    }
    /**
     * 选择一个节点并将光标聚焦在它的开始或结束位置
     * @param {Object} node
     * @param {Bool} begin - 与range.collapse()方法的参数作用一致，true表示开始位置，false表示结束位置
     */
    selection.selectNodeAndCollapse = function(node, begin){
        var sel = win.getSelection(),
            range = doc.createRange();
        range.selectNodeContents(node);
        range.collapse(begin);
        sel.removeAllRanges();
        sel.addRange(range);
    }


    /**
     ********
     * editor: 编辑器相关操作
     ********
     */
    var editor = QMUIEditor.editor =  doc.querySelector('#QMUIEditor');

    Object.defineProperties(editor, {
        /**
         * 设置为只读状态
         */
        'setReadOnly': {
            get: function(){
                return function(){
                    this.setAttribute('contenteditable', 'false');
                }
            }
        },
        /**
         * 清空编辑器内容还原为最初状态
         */
        'clear': {
            get: function(){
                return function(){
                    this.innerHTML = '<' + config.defaultParagraph + '><br></' + config.defaultParagraph + '>';
                }
            }
        },
        /**
         * 判断编辑器是否被聚焦
         */
        'isFocused': {
            get: function(){
                return function(){
                    return this === doc.activeElement;
                }
            }
        },
        /**
         * 检查编辑器的内容是否为空，来决定是否出现placeholder
         * 遵循以下规则：
         * 若存在具有特殊视觉表现的标签，如标题、列表、引用、图片等，则返回 false
         * 当innerText长度为1且br个数为1，相当于当前编辑器只有一个空的段落，返回 true
         * 判断innerText的长度，这里不判断trim后的情况，因为有时候空格可能是用户故意输入的
         */
        'isEmpty': {
            get: function(){
                return function(){
                    if(this.querySelectorAll('img, iframe, canvas, object, blockquote, ul, ' + CONST.titleTagArr.join(',')).length > 0){
                        return false;
                    }
                    if((this.innerText.length === 1) && (this.innerText.length === this.querySelectorAll('br').length)){
                        return true;
                    }
                    return this.innerText.length <= 0;
                }
            }
        },
        /**
         * 设置placeholder文案
         * @param {String} text
         */
        'setPlaceholder': {
            get: function(){
                return function(text){
                    this.setAttribute('placeholder', text);
                }
            }
        },
        /**
         * 切换placeholder的显示
         */
        'togglePlaceholder': {
            get: function(){
                return function(){
                    if(this.isEmpty()){
                        this.classList.add('QMUIEditor_empty');
                    }else{
                        this.classList.remove('QMUIEditor_empty');
                    }
                }
            }
        },
        /**
         * 聚焦并将光标置于编辑器最前面的位置
         */
        'focusEditorAtStartSelection': {
            get: function(){
                return function(){
                    selection.selectNodeAndCollapse(this, true);
                    this.focus();
                }
            }
        },
        /**
         * 使编辑器失焦
         * 需要把选区清除掉，否则在web上编辑器之外的地方的操作有可能会影响到编辑器
         */
        'blurfocus': {
            get: function(){
                return function(){
                    this.blur();
                    win.getSelection().removeAllRanges();
                }
            }
        },
        /**
         * 聚焦编辑器并将光标恢复到上一次backup的位置
         * 如果没有backup过则将光标移动到编辑器的最末尾
         */
        'focusEditorAtBackupSelection': {
            get: function(){
                return function(){
                    if(selection.current.startContainer) {
                        selection.restore();
                    }else{
                        selection.selectNodeAndCollapse(this, false);
                    }
                    this.focus();
                }
            }
        },
        /**
         * 调整页面的scrollTop，保证光标聚焦的地方可见且不会紧贴window下边缘
         * TODO:  禁止webview的滚动然后再调整
         */
        'updateScrollY': {
            get: function(){
                return function(){
                    var caretTopToWindowTop = selection.getYCoods().top,
                        caretBottomToWindowTop = selection.getYCoods().bottom,
                        windowHeight = win.innerHeight,
                        scrollTop = win.scrollY,
                        comfortableOffset = 20,
                        offset = 0;
                    if(caretBottomToWindowTop + comfortableOffset >= windowHeight){
                        offset = caretBottomToWindowTop - windowHeight + comfortableOffset;
                        win.scrollTo(0, scrollTop + offset);
                    }
                }
            }
        },
        /**
         * 设置编辑器的内容
         * @param {String} contents.
         */
        'setHtml': {
            get: function(){
                return function(contents){
                    this.innerHTML = contents;
                    this.togglePlaceholder();
                }
            }
        },
        /**
         * 获取编辑器的纯文本内容
         * @param {String} html.
         */
        'getPlainContent': {
            get: function(){
                return function(){
                    var plainContent = this.innerText;
                    config.onPushNative('updatePlainContent', plainContent);
                    return this.innerText;
                }
            }
        },
        /**
         * 返回完整的innerHTML
         */
        'getHtml': {
            get: function(){
                return function(){
                    var html = this.innerHTML;
                    config.onPushNative('updateContent', html);
                    return html;
                }
            }
        },
        /**
         * 返回处理过的html，目前会处理的有:
         * 1 检查通过insertDomHtml插入的iframe，并将其转换成html
         */
        'getTransformedContent': {
            get: function(){
                return function(){
                    var html = this.innerHTML,
                        shadowContainer = doc.createElement('div');
                    shadowContainer.innerHTML = html;
                    for(var ifrClass in win.insertedIframes){
                        if(ifrClass){
                            var ifrHtml = convertDomToHtmlString(shadowContainer.querySelector('.' + ifrClass));
                            html = html.replace(ifrHtml, win.insertedIframes[ifrClass]);
                        }
                    }
                    config.onPushNative('updateTransformedContent', html);
                    return html;
                }
            }
        },
        /**
         * 图片点击事件，会将图片的src和data-addition传给native
         */
        'onImgClick': {
            get: function(){
                return function(e){
                    var img = e.target,
                        src = img.getAttribute('src'),
                        addition = img.getAttribute('qmui-img-addition'),
                        data = {
                            'src': encodeURI(src),
                            'addition': addition?addition:''
                        };
                    config.onPushNative('onImgClick', JSON.stringify(data));
                }
            }
        },
        /**
         * iframe点击事件，会将iframe上绑定的数据传给java层
         */
        'onIframeClick': {
            get: function(){
                return function(){
                    config.onPushNative('onIframeClick', addition);
                }
            }
        }
    });
    /**
     * 监听编辑器内的img的点击并回调数据给app
     * 可以按需要增加其他类型的事件回调
     */
    editor.addEventListener('click', function(e){
        if(e.target.nodeName.toLowerCase() == 'img'){
            this.onImgClick(e);
        }
    });
    /**
     * 监听编辑器聚焦／失焦状态并回调
     */
    editor.addEventListener('focus', function(e){
        config.onPushNative('editorFocused', "true");
    });
    editor.addEventListener('blur', function(e){
        config.onPushNative('editorFocused', "false");
    });
    /*
     * 监听输入事件
     */
    editor.addEventListener('input', function(){
        // 保证编辑器首行有一个默认段落标签
        if(editor.isEmpty()){
            this.clear();
            selection.selectNodeAndCollapse(editor.firstChild, true);
        }
        // 更新placeholder的显示状态
        this.togglePlaceholder();
        // 更新页面scrollTop让光标位置不会那么紧贴边缘
        this.updateScrollY();
        // 回调页面内容给app
        this.getHtml();
        this.getPlainContent();
        this.getTransformedContent();
        // 避免在blockquote里生成段落
        edit.skipParasInBlockquote();
        // 从一个空blockquote里面按回车，理想情况是取消blockquote，像从一个空的li里面回车一样
        edit.enterInBlockquote();
        // 删掉多余的font标签
        edit.removeUnnecessaryFontTag();
        // 强制让列表被段落包住
        edit.enforceListInsidePara();
    });


    /**
     ********
     * 格式编辑方法
     ********
     */
    var edit = QMUIEditor.edit = {};
    /**
     * 编辑器里生成font标签后设置列表
     * 如果
     * 1 font标签的parent的唯一子node就是这个font
     * 2 font标签的parent的子node只有这个font和一个br标签
     * 设置／取消列表的操作就会导致font标签消失，系统会主动做类似unwrap的行为
     * 这里在需要的时候为font标签前（或者后）加一个任意的字符node，就能避免这个行为
     * 步骤：
     * 保存选区，加字符，恢复选区－－执行需要的操作－－保存选区，去掉字符，恢复选区
     * 选区的保存和恢复是必要的，因为任意dom的增减都会导致选区丢失
     *
     * TODO：目前的做法比较粗糙，待优化
     */
    function addSpaceForSavingFontTags(){
        selection.backup();
        var fonts = doc.querySelectorAll('font');
        if(fonts.length > 0){
            for(var i = 0;i <fonts.length; i++){
                var font = fonts[i],
                    fontParent = font.parentNode;
                if((font == fontParent.firstChild && font == fontParent.lastChild) ||
                    (font == fontParent.firstChild && fontParent.lastChild.nodeName.toLowerCase() == 'br' && fontParent.childNodes.length == 2)){
                    var spaceTag = doc.createElement('i');
                    spaceTag.className = 'QMUIEditor_func_space';
                    spaceTag.innerHTML = '&nbsp;';
                    spaceTag.prependBefore(font);
                }
            }
            selection.restore();
        }
    }
    function removeSpaceForSavingFontTags(){
        selection.backup();
        var spaces = doc.querySelectorAll('.QMUIEditor_func_space');
        if(spaces.length > 0){
            for(var i = 0; i < spaces.length; i++){
                spaces[i].remove();
            }
            selection.restore();
        }
    }
    /**
     * 调用系统命令进行退格删除
     */
    edit.delete = function() {
        doCommand('delete');
    }
    /**
     * 撤销最近执行的命令
     */
    edit.undo = function() {
        doCommand('undo');
    }
    /**
     * 重做被撤销的操作
     */
    edit.redo = function() {
        doCommand('redo');
    }
    /**
     * 插入一个包含多个子node的dom，该dom可以被完整删除。
     * 通过插入iframe实现。
     * 插入的dom的display样式会强制改为inline-block。因为如果插入的是一个display:block的元素，会影响编辑过程中的段落合并，导致段落格式错乱；而如果是一个display:inline则会导致一些样式失效。
     * @beta
     * @param {String} domHtml - dom的字符串表示
     * @param {String} clickData - 附加数据，可用于点击时传值或其它场景
     * 
     * '<span>test</span>' right
     * '<span>test</span><span>test</span>'wrong
     */
    edit.insertDomHtml = function(domHtml, clickData) {
        var dom = convertHtmlStringToDom(domHtml);
        if(!dom) return;

        // 1 获取该dom插入后的尺寸、margin、display等外部样式
        var shadowEditor;
        if(this.shadowEditor){
            shadowEditor = this.shadowEditor
        }else{
            shadowEditor = doc.createElement('div');
            shadowEditor.className = 'QMUIEditor';
            // shadowEditor.style.cssText = 'position:absolute;bottom:0;left:0;right:0;min-height:0';
            shadowEditor.style.cssText = 'position:absolute;top:-9999px;left:0;right:0;';
            doc.body.appendChild(shadowEditor);
            this.shadowEditor = shadowEditor;
        }
        shadowEditor.innerHTML = domHtml;
        var shadowEditorFirstChild = shadowEditor.firstChild;
        shadowEditorFirstChild.style.display = 'inline-block';
        var shadowEditorFirstChildRect = shadowEditorFirstChild.getBoundingClientRect(),
            outerWidth, outerHeight;
        outerWidth = shadowEditorFirstChildRect.width;
        if(['img', 'video'].indexOf(shadowEditorFirstChild.tagName.toLowerCase()) >= 0){
            if(!shadowEditorFirstChild.width || !shadowEditorFirstChild.height){
                console.warn(CONST.logTag.warning + '插入的dom如果是video或者img标签，请加上width和height属性。');
                outerHeight = shadowEditorFirstChildRect.height;
            }else{
                outerHeight = (outerWidth/shadowEditorFirstChild.width)*shadowEditorFirstChild.height;
            }
        }else{
            outerHeight = shadowEditorFirstChildRect.height;
        }
        var domOuterStyle = getDomStyle(win, shadowEditorFirstChild),
            outerStyleCssText = 'margin:' + domOuterStyle['margin'] + ';'
                                + 'border-radius:' + domOuterStyle['border-radius'] + ';'
                                + 'box-shadow:' + domOuterStyle['box-shadow'] + ';'
                                + 'vertical-align:' + domOuterStyle['vertical-align'] + ';'
                                + 'width:' + outerWidth + 'px;'
                                + 'height:' + outerHeight + 'px;';
        if(domOuterStyle['display'] == 'block'){
            console.warn(CONST.logTag.warning + '插入的dom的display样式会强制改为inline-block。');
        }
        outerStyleCssText += 'display: inline-block;';
        shadowEditorFirstChild.style.cssText = 'display:inline-block;margin:0;box-shadow:none;vertical-align: top;';
        
        // 2 创建iframe用来代替dom插入到编辑器中
        var ifrClass = 'ifr_' + new Date().getTime(),
            ifrHtmlString = '<iframe class="func_iframe ' + ifrClass + '" style="' + outerStyleCssText + '"></iframe>';
        win.insertedIframes = win.insertedIframes || {};
        win.insertedIframes[ifrClass] = domHtml;
        this.insertHtml(ifrHtmlString);
        
        // 3 通过js插入dom 
        var ifr = doc.querySelector('.' + ifrClass),
            ifrDoc = ifr.contentDocument,
            ifrWin = ifr.contentWindow,
            ifrHd = ifrDoc.head,
            ifrBd = ifrDoc.body;
        ifrBd.className = 'func_iframe_body';
        if (clickData) ifrDoc.onclick = editor.onIframeClick(clickData);
        win.editorStylesArr = win.editorStylesArr || doc.querySelectorAll('link[rel="stylesheet"]');
        for(var i = 0; i < win.editorStylesArr.length; i++){
            ifrHd.appendChild(win.editorStylesArr[i].cloneNode());
        }
        ifrBd.innerHTML = shadowEditor.innerHTML;
    }

    /**
     * 原生的insert命令需要页面存在selection时候才可以进行，与是否被聚焦无关。这里加个通用方法做下判断确保有一个selection
     * @param {String} content - 插入的内容
     * @param {Bool} onlyText - 是否用纯文本
     */
    edit.insert = function(content, onlyText){
        if(selection.current.startContainer){
            selection.restore();
        }else{
            selection.selectNodeAndCollapse(editor, false);
        }
        console.log('test selection.current')
        console.log(selection.current)
        doc.execCommand(onlyText?'insertText':'insertHTML', false, content);
    }
    /**
     * 插入html片段，用的是默认的系统命令
     */
    edit.insertHtml = function(html){
        this.insert(html);
    }
    /**
     * 插入文本，用的是默认的系统命令
     */
    edit.insertText = function(text){
        this.insert(text, true);
    }
    /**
     * 插入图片
     * @param {Object} img
     * @param {String} img.src
     * @param {String} img.alt
     * @param {Int} img.w - 图片原始高度
     * @param {Int} img.h - 图片原始高度
     * @param {String} img.addition - 附加数据
     */
    edit.insertImage = function(img){
        if(!img.src){
            return;
        }
        var imgHtml = '<img src="'+img.src+'" alt="'+img.alt+'" qmui-img-addition="'+img.addition+'" ' + (img.w?('width="' + img.w + '"'):('')) + ' ' + (img.h?('height="' + img.w + '"'):('')) + ' />';
        this.insert(imgHtml);
        // 下一帧重画后重新调整y轴滚动条位置，否则不准
        win.requestAnimationFrame(function(){
            editor.updateScrollY();
        });
    }
    /**
     * 插入一个空的段落，并分割当前光标所在段落
     */
    edit.insertParaAndBreakCurrentBlock = function(){
        this.insertHtml('<' + config.defaultParagraph +  '><br></' + config.defaultParagraph +  '>');
    }
    /**
     * 插入超链接
     * @param {String} url
     * @param {String} text
     */
    edit.insertLink = function(url, text){
        if(url && url.length > 0){
            var t = text || url;
            edit.insertHtml('<a href="' + url + '">' + t + '</a>');  
        }
    }
    /**
     * 将选中的文字变成超链接
     * @param {String} url
     */
    edit.createLink = function(url){
        if(selection.current.length > 0){
            doCommand('createLink', url);
        }else{
            edit.insertLink(url);
        }
    }
    /**
     * 取消超链接
     */
    edit.unlink = function(url){
        doCommand('unlink', false, null);
    }
    /**
     * 如果当前光标选中了超链接，则可以更新它，类似读书里的@人和话题
     * selection里会当光标处在超链接中时通知java层
     * @param {Object} newLink
     * @param {String} newLink.url
     * @param {String} newLink.text
     */
    edit.updateLink = function(newLink){
        if(selection.current.link){
            var link = selection.current.link;
            link.innerText = newLink.text || link.innerText;
            link.href = newLink.url || link.href;
        }
    }
    /**
     * 设置斜体
     */
    edit.setItalic = function(){
        doCommand('italic');
        edit.removeSpanGeneratedBySystem();
        state.push();
    }
    /**
     * 设置下划线
     */
    edit.setUnderline = function(){
        doCommand('Underline');
        edit.removeSpanGeneratedBySystem();
        state.push();
    }
    /**
     * 设置删除线
     */
    edit.setStrikeThrough = function(){
        doCommand('StrikeThrough');
        edit.removeSpanGeneratedBySystem();
        state.push();
    }
    /**
     * 加粗
     * 在没有选中任何文字时，设置加粗、字体颜色、字号大小有可能不会触发selectionchange事件，不能及时提交光标所在位置的状态。所以这里要主动push一次
     * 
     */
    edit.setBold = function(){
        doCommand('bold');
        edit.removeSpanGeneratedBySystem();
        state.push();
    }
    /**
     * 设置文字颜色，生成标签 <font color="color">xxxxx</font>
     * @param {string} color eg: #ff0000
     */
    edit.setTextColor = function(color){
        doCommand('forecolor', color);
        edit.removeSpanGeneratedBySystem();
        state.push();
    }
    /**
     * 设置字号，生成标签 <font size="size">xxxxx</font>
     * @param {int} size 字号的档次，三个值看config里的设定
     * https://developer.mozilla.org/zh-CN/docs/Web/API/doc/execCommand
     * 可以在css里重置每个档次的font标签的实际像素值，例如
     * <style>font[size="4"]{19px}</style>
     * 不过size对应的允许的像素值请看下表
     * <font size=”?“>  CSS font-size; ?
     *      1               太小，忽略
     *      2               ~14px
     *      3               15px-16px
     *      4               17px-20px
     *      5               21px-27px
     *      6               28px-39px
     *      7               40px~
     */
    edit.setFontSize = function(size){
        doc.execCommand("fontSize", false, size);
        this.removeSpanGeneratedBySystem();
        this.removeUnnecessaryFontTag();
        state.push();
    }
    edit.setLargeFontSize = function(){
        this.setFontSize(config.fontSizeLarge);
    }
    edit.setBigFontSize = function(){
        this.setFontSize(config.fontSizeBig);
    }
    edit.setNormalFontSize = function(){
        this.setFontSize(config.fontSizeNormal);
    }
    edit.setSmallFontSize = function(){
        this.setFontSize(config.fontSizeSmall);
    }
    /**
     * 设置／取消居中
     */
    edit.setJustifyCenter = function(){
        if (getCommandState('justifyCenter')) {
            doCommand('justifyLeft');
        }else{
            doCommand('justifyCenter');
        }
        this.removeSpanGeneratedBySystem();
        state.push();
    }
    /**
     * 设置／取消两端对齐
     */
    edit.setJustifyFull = function(){
        if (getCommandState('justifyFull')) {
            doCommand('justifyLeft');
        }else{
            doCommand('justifyFull');
        }
        this.removeSpanGeneratedBySystem();
        state.push();
    }
    /**
     * 设置／取消右对齐
     */
    edit.setJustifyRight = function(){
        if (getCommandState('justifyRight')) {
            doCommand('justifyLeft');
        }else{
            doCommand('justifyRight');
        }
        this.removeSpanGeneratedBySystem();
        state.push();
    }
    /**
     * 设置左对齐
     */
    edit.setJustifyLeft = function(){
        doCommand('justifyLeft');
        this.removeSpanGeneratedBySystem();
        state.push();
    }
    /**
     * 判断标签名是否属于标题
     * @param {String} tagName
     * @return {boolean}
     */
    edit.isHeadingFormat = function(tagName){
        return CONST.titleTagArr.indexOf(tagName) >= 0;
    }
    /**
     * 设置标题。标题不与断落、列表和引用标签共存，会互相覆盖
     * @param {String} titleTag  h1/h2/h3/h4/h5/h6
     */
    edit.setTitle = function(titleTag) {
        var formatBlock = getCommandValue('formatBlock');
        if (formatBlock === 'blockquote') {
            doCommand('formatBlock', '<' + config.defaultParagraph + '>')
        }
        if (formatBlock.length > 0 && formatBlock.toLowerCase() === titleTag) {
            doCommand('formatBlock', '<' + config.defaultParagraph + '>');
        } else {
            doCommand('formatBlock', '<' + titleTag + '>');
        }
        if (getCommandState('insertUnorderedList')) {
            doCommand('InsertUnorderedList');
        }
        if (getCommandState('insertOrderedList')) {
            doCommand('insertOrderedList');
        }
        this.removeSpanGeneratedBySystem();
        this.removeUnnecessaryFontTag();
    }
    /**
     * 设置引用格式，与title互斥
     */
    edit.setBlockquote = function() {
        var formatBlock = getCommandValue('formatBlock');
        if (this.isHeadingFormat(formatBlock)) {
            doCommand('formatBlock', '<' + config.defaultParagraph + '>');
        }
        if (formatBlock == 'blockquote') {
            doCommand('formatBlock', '<' + config.defaultParagraph + '>')
        } else {
            doCommand('formatBlock', '<blockquote>');
        }
        this.removeSpanGeneratedBySystem();
    }
    /**
     * 设置无序列表
     */
    edit.setUnorderedList = function() {
        addSpaceForSavingFontTags();
        var formatBlock = getCommandValue('formatBlock');
        if (this.isHeadingFormat(formatBlock)) {
            doCommand('formatBlock', '<' + config.defaultParagraph + '>');
        }
        doCommand('InsertUnorderedList');
        this.removeSpanGeneratedBySystem();
        removeSpaceForSavingFontTags();
    }
    /**
     * 设置顺序列表
     */
    edit.setOrderedList = function() {
        addSpaceForSavingFontTags();
        var formatBlock = getCommandValue('formatBlock');
        if (this.isHeadingFormat(formatBlock)) {
            doCommand('formatBlock', '<' + config.defaultParagraph + '>');
        }
        doCommand('InsertOrderedList');
        this.removeSpanGeneratedBySystem();
        removeSpaceForSavingFontTags();
    }
    /**
     * 去掉系统自动生成的span
     * 系统有一套生成span的逻辑，目的是用span来保留一些标签在操作过后样式的变化
     * 例如设置列表
     * <style>li{color:red;}</style>
     * <编辑器><div><ul><li>asgasg</li></ul></div></编辑器>
     * 这时候li包住的文字自然有color:red的样式
     * 当取消列表，理想中它应该变成<div>asgasg</div>，但系统会觉得原本拥有color:red这个样式的文字还要保持颜色不变，所以系统会用一个span把文字包住并把要保留下来的样式加到span上，于是最终就变成：
     * <div><span style="color: red;">asgasg</span><br></div>
     * 我们希望样式由css文件来控制，所以这里把span标签都unwrap掉
     */
    edit.removeSpanGeneratedBySystem = function(){
        // 去掉系统自动生成的span
        var spans = QMUIEditor.editor.querySelectorAll('span[style]');
        if(spans.length > 0){
            selection.backup();
            for(var i = 0; i<spans.length; i++){
                var span = spans[i];
                if(span.parentNode == editor){
                    span.wrapBy(config.defaultParagraph);
                }
                spans[i].unwrap();
            }
            selection.restore();
        }
    }
    /**
     * unwrap掉不需要的font标签，包括
     * 1 size＝3，也就是正常大小的字体
     * 2 按需要屏蔽掉标题标签里的字体设置
     * 标题自带大字号，在标题里再另外单独设置字号大小，无论视觉上还是操作逻辑上个人觉得都有点奇怪
     * 而且由于标题肯定是会自带不同的（一般都是比较大的）字号，这会影响对光标当前位置字号大小的查询，当然这是因为系统execCommand命令返回值的一些局限，目前还没有很好的解决方法
     * 所以默认会把title里的font标签都屏蔽掉去掉
     * 假设真的想同时用到这两种格式设置，可以设置config里的ignoreFontsInTitle为false即可
     * 同时记得在java里根据状态回调修改相应的java逻辑
     *
     * 注意：
     * 该方法在各种setTitle和setFontSize里调用之外，也会在编辑器的input回调里调用
     * 因为目前setFontSize用的是execCommand，如果没有选中内容，该方法不会生成新标签，而是会做一种类似标记的操作
     * 执行setFontSize之后接着的输入行为，才会生成font标签
     */
    edit.removeUnnecessaryFontTag = function(){
        // 相当于css选择器：font[size="3"], h1 font[size],h2 font[size],h3 font[size],h4 font[size],h5 font[size],h6 font[size]
        var fontTagQueryStr = 'font[size="3"]';
        if(config.ignoreFontsInTitle){
            for(var i = 0; i < CONST.titleTagArr.length; i++){
                fontTagQueryStr += ',' + CONST.titleTagArr[i] + ' font[size]';
            }
        }
        var targetTags = doc.querySelectorAll(fontTagQueryStr);
        if(targetTags.length > 0){
            selection.backup();
            for(var i = 0; i < targetTags.length; i++){
                var font = targetTags[i];
                font.unwrap();
            }
            selection.restore();
        }
    }
    /**
     * 避免在blockquote里产生段落
     * 重现路径:在blockquote里设置列表,回车取消列表,这时会产生一个段落元素。理想的情况应该在这时候光标跳出blockquote
     */
    edit.skipParasInBlockquote = function(){
        var targetParas = doc.querySelectorAll('blockquote ' + config.defaultParagraph + ':last-child, blockquote ' + config.defaultParagraph + ':only-child');
        if(targetParas.length > 0){
            selection.backup();
            for(var i = 0; i < targetParas.length; i++){
                var para = targetParas[i],
                    blockquote = para.parentNode;
                para.appendAfter(blockquote);
                if(blockquote.childNodes.length == 0){
                    blockquote.remove();
                }
            }
            selection.restore();
        }
    }
    /**
     * 从一个空blockquote里面按回车，离线情况是取消blockquote，像从一个空的li里面回车一样
     */
    edit.enterInBlockquote = function(){
        var target = doc.querySelector('blockquote + blockquote:last-of-type');
        if(!target) return;
        var targetPrev = target.previousSibling;
        if(target.childNodes.length == 1 && target.firstChild.nodeName.toLowerCase() == 'br' && targetPrev.childNodes.length == 1 && targetPrev.firstChild.nodeName.toLowerCase() == 'br'){
            var newPara = doc.createElement(config.defaultParagraph);
            newPara.innerHTML = '<br>';
            newPara.prependBefore(targetPrev);
            target.remove();
            targetPrev.remove();
            selection.selectNodeAndCollapse(newPara, false);
        }
    }
    /**
     * 强制让列表被段落包裹住
     */
    edit.enforceListInsidePara = function(){
        var list = doc.querySelectorAll('.QMUIEditor > ul, .QMUIEditor > ol');
        if(list.length > 0){
            selection.backup();
            for(var i = 0; i < list.length; i++){
                list[i].wrapBy(config.defaultParagraph);
            }
            selection.restore();
        }
    }

    /*
     ********
     * 初始化:
     * 设置默认首行段落标签
     * 监听光标选区变化事件，完成：
     * 1 editor.togglePlaceholder();
     * 2 回调当前光标选区所在位置的标志给app，具体标志见 state.map
     * 3 及时备份选区
     ********
     */
    QMUIEditor.init = function(conf){
        if(typeof conf == "object"){
            for(var a in conf){
                config[a] = conf[a];
            }
        }
        doCommand('defaultParagraphSeparator', config.defaultParagraph);
        editor.clear();
        doc.addEventListener('selectionchange', function(){
            editor.togglePlaceholder();
            console.log('test selection.current after selectionchange')
            console.log(selection.current)
            selection.backup();
            state.push();
        });
    }

})(window, document);

// 初始化
QMUIEditor.init();
