/**
 * tools
 */

/**
 * 拓展Node以更加方便地进行一些操作
 */
if (typeof Node != 'undefined') {
    /**
     * 寻找该节点的标签名必须为指定tagName的祖先元素
     * @param  {String | Array} tagName  指定的标签名
     * @return {HTMLElement} DOM元素，如果不存在，返回null
     */
    Node.prototype.getParentElementByTagName = function(tagName){
        if(Array.isArray(tagName)){
            var targetNode = [];
            for(var i = 0; i < tagName.length; i++){
                var node = this.getParentElementByTagName(tagName[i]);
                if(node) targetNode.push(node);
            }
            return targetNode.length > 0 ? targetNode : null;
        }else{
            var targetNodeName = tagName.toLowerCase();
            var targetNode = this.parentElement;
            while (targetNode && targetNode.tagName.toLowerCase() != targetNodeName) {
                targetNode = targetNode.parentElement;
            }
            return targetNode;
        }
    }
    /**
     * 移除该节点
     */
    Node.prototype.remove = function(){
        if (this.parentNode) {
            this.parentNode.removeChild(this);
        }
    }
    /**
     * 为该包一个标签
     * @param  {string} tagName  指定的标签名
     */
    Node.prototype.wrapBy = function(tagName){
        var wrapDom = document.createElement(tagName);
        wrapDom.prependBefore(this);
        wrapDom.appendChild(this);
    }
    /**
     * 将该节点插到目标节点后面
     * @param  {Node} targetNode  目标节点
     */
    Node.prototype.appendAfter = function(targetNode){
        var parentEle = targetNode.parentElement;

        if (parentEle.lastChild === targetNode){
            parentEle.appendChild(this);
        } else {
            parentEle.insertBefore(this,targetNode.nextElementSibling);
        }
    }
    /**
     * 将该节点插到目标节点前面
     * @param  {Node} targetNode  目标节点
     */
    HTMLElement.prototype.prependBefore = function(targetNode){
        targetNode.parentElement.insertBefore(this, targetNode);
    }
}

/**
 * 拓展HTMLElement以更加方便地进行一些操作
 */
if (typeof HTMLElement != 'undefined') {
    HTMLElement.prototype.unwrap = function(){
        /**
         * 将element的所有子node取出组成新的fragment并替换掉该element inspired by medium-editor
         */
        var self = this;
        if(!self.childNodes.length){
            self.parentNode.removeChild(self);
            return;
        }
        var fragment = document.createDocumentFragment(),
            nodes = Array.prototype.slice.call(self.childNodes);
        for (var i = 0; i < nodes.length; i++) {
            fragment.appendChild(nodes[i]);
        }
        self.parentNode.replaceChild(fragment, self);
    }
}

/**
 * 根据html字符串创建dom，该字符串必须只表示一个dom
 * @param {String} htmlString - 要创建的dom的字符串
 * @return {Object} 返回根据html字符串创建的dom
 */
function convertHtmlStringToDom(htmlString){
    var outer = document.createElement('div'),
        dom = null;
    outer.innerHTML = htmlString;
    if(outer.childNodes.length == 1 && outer.firstChild.nodeType == 1){
        dom = outer.firstChild;
    }
    return dom;
}

/**
 * 将dom转化成html字符串
 * @param {Object} dom - 待转化的dom
 * @return {String} 目标dom的字符串表现
 */
function convertDomToHtmlString(dom){
    if(dom){
        var cloned = dom.cloneNode(true),
            outer = document.createElement('div');
        outer.appendChild(cloned);
        return outer.innerHTML;
    }
}

/**
 * 判断对象是否为一个dom node
 * @param  {Object} obj
 * @return {Boolean}
 */
function isNode(obj){
    return (
        typeof Node === "object" ? obj instanceof Node : 
            obj && typeof obj === "object" && typeof obj.nodeType === "number" && typeof obj.nodeName==="string"
    );
}

/**
 * 判断对象是否为一个dom element
 * @param  {Object} obj
 * @return {Boolean}
 */
function isElement(obj){
    return (
        typeof HTMLElement === "object" ? obj instanceof HTMLElement : 
            obj && typeof obj === "object" && obj !== null && obj.nodeType === 1 && typeof obj.nodeName==="string"
    );
}

/**
 * rgb2hex 将颜色的rgb值转化为hex的表示
 * 
 * @param  {(string|number)} red
 * @param  {(string|number)} green
 * @param  {(string|number)} blue
 * @return {string} 用hex表示的颜色值
 * 
 */
function rgb2hex(red, green, blue) {
    var rgb = blue | (green << 8) | (red << 16);
    return '#' + (0x1000000 + rgb).toString(16).slice(1)
}

/**
 * 获取指定样式的值
 * @param {Object} win - window对象
 * @param {Object} dom
 * @param {String} pseudoClass - 伪类名字
 * @return {CSSStyleDeclaration} 目标元素的样式
 */
function getDomStyle(win, dom, pseudoClass){
    if(win.getComputedStyle && dom){
        return win.getComputedStyle(dom, pseudoClass || null);
    }
    return null;
}

/**
 * 获取元素的contentWidth
 * @param {Object} dom
 * @return {Number} 元素的内容宽度
 */
function getDomContentWidth(dom){
    return dom ? (dom.offsetWidth
            - parseInt(getDomStyle(window, dom)['padding-left'])
            - parseInt(getDomStyle(window, dom)['padding-right'])
            - parseInt(getDomStyle(window, dom)['border-left-width'])
            - parseInt(getDomStyle(window, dom)['border-right-width'])) : null;
}

/**
 * requestAnimationFrame兼容封装
 */
(function() {
    var lastTime = 0;
    var vendors = ['ms', 'moz', 'webkit', 'o'];
    for(var x = 0; x < vendors.length && !window.requestAnimationFrame; ++x) {
        window.requestAnimationFrame = window[vendors[x]+'RequestAnimationFrame'];
        window.cancelAnimationFrame = window[vendors[x]+'CancelAnimationFrame']
                                   || window[vendors[x]+'CancelRequestAnimationFrame'];
    }

    if (!window.requestAnimationFrame)
        window.requestAnimationFrame = function(callback, element) {
            var currTime = new Date().getTime();
            var timeToCall = Math.max(0, 16 - (currTime - lastTime));
            var id = window.setTimeout(function() { callback(currTime + timeToCall); },
              timeToCall);
            lastTime = currTime + timeToCall;
            return id;
        };

    if (!window.cancelAnimationFrame)
        window.cancelAnimationFrame = function(id) {
            clearTimeout(id);
        };
}());

/**
 * 系统命令相关的方法
 */
function getCommandValue(command){
    return document.queryCommandValue(command);
}
function getCommandState(command){
    return document.queryCommandState(command) 
}
function doCommand(command, arg){
    document.execCommand(command, false, arg?arg:null);
}
function appendStyleLinkTag(url){
    var tag =  document.createElement('link');
    tag.rel = 'stylesheet';
    tag.href = url;
    document.querySelector('head').appendChild(tag);
}
function appendScriptTag(url){
    var tag =  document.createElement('script');
    tag.src = url;
    document.querySelector('body').appendChild(tag);
}